/*
 *	Copyright (c) 2000 Rinet Corp., Novosibirsk, Russia
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 *
 * THIS SOURCE CODE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/time.h>

/* Version 1 Header Format */
typedef	struct header_v1 {
	u_short		version;	/* version number=1 */
	u_short		counter;	/* number of exported flows (1-24) */
	time_t		sysuptime;	/* milliseconds since router booted */
	struct timespec	currtime;	/* current time since UTC */
} HEADER_V1;

/* Version 1 Flow Record Format */
typedef	struct flow_record_v1 {
	in_addr_t	src_addr;	/* source IP address */
	in_addr_t	dst_addr;	/* destination IP address */
	in_addr_t	nexthop;	/* next hop router's IP address */
	u_short		ifin;		/* input interface's SNMP index */
	u_short		ifout;		/* output interface's SNMP index */
	u_long		dpkts;		/* packets in the flow */
	u_long		doctets;	/* total number of L3 bytes */
	time_t		firsttime;	/* sysuptime at start of flow */
	time_t		lasttime;	/* sysuptime at last packet of flow */
	u_short		src_port;	/* source port number */
	u_short		dst_port;	/* destination port number */
	u_short		pad1;		/* unused (zero) bytes */
	u_char		proto;		/* IP protocol */
	u_char		tos;		/* type of service */
	u_char		flags;		/* cumulative OR of TCP flags */
	u_char		pad2;		/* unused (zero) byte */
	u_short		pad3;		/* unused (zero) byte */
	u_char		reserved[8];	/* unused (zero) bytes */
} FLOW_RECORD_V1;

/* Version 5 Header Format */
typedef	struct header_v5 {
	u_short		version;	/* version number=5 */
	u_short		counter;	/* number of exported flows (1-30) */
	time_t		sysuptime;	/* milliseconds since router booted */
	struct timespec	currtime;	/* current time since UTC */
	u_long		sequence;	/* sequence counter of total flows */
	u_char		engine_type;	/* switching engine type (RP,VIP) */
	u_char		engine_id;	/* switching engine slot number */
} HEADER_V5;

/* Version 5 Flow Record Format */
typedef	struct flow_record_v5 {
	in_addr_t	src_addr;	/* source IP address */
	in_addr_t	dst_addr;	/* destination IP address */
	in_addr_t	nexthop;	/* next hop router's IP address */
	u_short		ifin;		/* input interface's SNMP index */
	u_short		ifout;		/* output interface's SNMP index */
	u_long		dpkts;		/* packets in the flow */
	u_long		doctets;	/* total number of L3 bytes */
	time_t		firsttime;	/* sysuptime at start of flow */
	time_t		lasttime;	/* sysuptime at last packet of flow */
	u_short		src_port;	/* source port number */
	u_short		dst_port;	/* destination port number */
	u_char		pad1;		/* unused (zero) byte */
	u_char		flags;		/* cumulative OR of TCP flags */
	u_char		proto;		/* IP protocol */
	u_char		tos;		/* type of service */
	u_short		dst_as;		/* AS of the destination */
	u_short		src_as;		/* AS of the source (origin of peer) */
	u_char		dst_mask;	/* dest address prefix mask bits */
	u_char		src_mask;	/* source address prefix mask bits */
	u_short		pad2;		/* unused (zero) bytes */
} FLOW_RECORD_V5;

/* Version 7 Header Format */
typedef	struct header_v7 {
	u_short		version;	/* version number=7 */
	u_short		counter;	/* number of exported flows (1-30) */
	time_t		sysuptime;	/* milliseconds since router booted */
	struct timespec	currtime;	/* current time since UTC */
	u_long		sequence;	/* sequence counter of total flows */
	u_long		reserved;	/* unused (zero) bytes */
} HEADER_V7;

/* Version 7 Flow Record Format */
typedef	struct flow_record_v7 {
	in_addr_t	src_addr;	/* source IP address */
	in_addr_t	dst_addr;	/* destination IP address */
	in_addr_t	nexthop;	/* next hop router's IP address */
	u_short		ifin;		/* input interface's SNMP index */
	u_short		ifout;		/* output interface's SNMP index */
	u_long		dpkts;		/* packets in the flow */
	u_long		doctets;	/* total number of L3 bytes */
	time_t		firsttime;	/* sysuptime at start of flow */
	time_t		lasttime;	/* sysuptime at last packet of flow */
	u_short		src_port;	/* source port number */
	u_short		dst_port;	/* destination port number */
	u_char		pad1;		/* unused (zero) byte */
	u_char		flags;		/* cumulative OR of TCP flags */
	u_char		proto;		/* IP protocol */
	u_char		tos;		/* type of service */
	u_long		src_as;		/* AS of the source (origin of peer) */
	u_long		dst_as;		/* AS of the destination */
	u_char		src_mask;	/* source address prefix mask bits */
	u_char		dst_mask;	/* dest address prefix mask bits */
	u_short		pad2;		/* unused (zero) bytes */
	in_addr_t	router_sc;	/* Router which is shortcut by switch */
} FLOW_RECORD_V7;
